local addon_ToCName, addon = ...;

LFG_TIMER_CLONE_FRAME_HEIGHT = 153;
LFG_TIMER_CLONE_FRAME_WIDTH = 300;

_G.LFGTimer = LibStub("AceAddon-3.0"):NewAddon(addon, addon_ToCName, "AceEvent-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("LFGTimer");
local LSM = LibStub("LibSharedMedia-3.0");

LSM:Register("Background", "Blizzard Chat Bubble", "Interface\\Tooltips\\ChatBubble-Background");

local db;
local defaults = {
	profile = {
		Frame = {
			Show = true,
			Type = "custom",
			Locked = false,
			Anchor = {
				Point = "CENTER",
				To = "UIParent",
				At = "CENTER",
				X = 0,
				Y = 0,
			},
			Queue = {
				States = {
					notQueued = true,
					proposal = true,
					inParty = true,
					queuedNoData = true,
					queued = true,
				},
				Types = {
					LookingForDungeon = true,
					RaidFinder = true,
					Scenarios = true,
					PvP = true,
					WorldPvP = true,
				},
				Data = {
					LookingForDungeon = {
						QueueTimes = {
							Tank = true,
							Healer = true,
							DPS = true,
							Self = true,
						},
						Party = {
							Tank = true,
							Healer = true,
							DPS = true,
							Total = true,
						},
					},
					RaidFinder = {
						QueueTimes = {
							Tank = true,
							Healer = true,
							DPS = true,
							Self = true,
						},
						Party = {
							Tank = true,
							Healer = true,
							DPS = true,
							Total = true,
						},
					},
					Scenarios = {
						QueueTimes = {
							DPS = true,
							Self = true,
						},
						Party = {
							DPS = true,
							Total = true,
						},
					},
					PvP = {
						TeamSize = true,
						WaitTime = true,
					},
					WorldPvP = {
						TimeLeft = true,
					},
				},
			},
			Backdrop = {
				Background = "Blizzard Tooltip",
				Border = "Blizzard Tooltip",
				BackgroundColor = {
					r = 0,
					g = 0,
					b = 0,
					a = 1,
				},
				BorderColor = {
					r = 1,
					g = 1,
					b = 1,
					a = 1,
				},
				Inset = 4,
				Tile = false,
				TileSize = 32,
				EdgeSize = 16,
			},
			Text = {
				Font = LSM:GetDefault("font"),
				Size = 12,
				Outline = " ",
				Color = {
					r = 1,
					g = 1,
					b = 1,
					a = 1,
				},
			},
		},
	},
};

function addon:OnInitialize()
	self.Vars = LibStub("AceDB-3.0"):New("LFGTimerDB", defaults);
	self.Vars.RegisterCallback(self, "OnProfileChanged", "UpdateDB");
	self.Vars.RegisterCallback(self, "OnProfileCopied", "UpdateDB");
	self.Vars.RegisterCallback(self, "OnProfileReset", "UpdateDB");
	
	db = self.Vars.profile;
	
	self:InitBlizOptions();
	self:InitQueueData();
	self:UpdateFrameShown();
end

function addon:OnEnable()
	-- LFG/LFD/LFR stuff
	self:RegisterEvent("LFG_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_ROLE_CHECK_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_UPDATE", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_FAILED", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_SUCCEEDED", "UpdateFrameShown");
	self:RegisterEvent("LFG_PROPOSAL_SHOW", "UpdateFrameShown");
	self:RegisterEvent("LFG_QUEUE_STATUS_UPDATE", "UpdateFrameShown");
	-- PvP status
	self:RegisterEvent("UPDATE_BATTLEFIELD_STATUS", "UpdateFrameShown");
	-- World PvP
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "UpdateFrameShown");
	self:RegisterEvent("ZONE_CHANGED", "UpdateFrameShown");
	self:RegisterEvent("BATTLEFIELD_MGR_QUEUE_REQUEST_RESPONSE", "UpdateFrameShown");
	self:RegisterEvent("BATTLEFIELD_MGR_EJECT_PENDING", "UpdateFrameShown");
	self:RegisterEvent("BATTLEFIELD_MGR_EJECTED", "UpdateFrameShown");
	self:RegisterEvent("BATTLEFIELD_MGR_QUEUE_INVITE", "UpdateFrameShown");
	self:RegisterEvent("BATTLEFIELD_MGR_ENTRY_INVITE", "UpdateFrameShown");
	self:RegisterEvent("BATTLEFIELD_MGR_ENTERED", "UpdateFrameShown");
end

function addon:OnDisable()
	self:UnregisterAllEvents();
end

function addon:UpdateDB(event, database)
	db = database.profile;
	addon:UpdateFrameArt();
	addon:UpdateFrameShown();
end

function addon:InitBlizOptions()
	local frame = CreateFrame("Frame", "LFGTimerBlizOptionsPanel", UIParent);
	frame.name = "LFG Timer";
	local button = CreateFrame("Button", "LFGTimerBlizOptionsPanelButton", frame, "UIPanelButtonTemplate");
	button.Text = button:GetFontString();
	button.Text:SetText("Configure");
	button:SetHeight(button.Text:GetStringHeight() + 10);
	button:SetWidth(button.Text:GetStringWidth() + 10);
	button:SetScript("OnClick", function()
		addon:OpenConfig()
	end);
	button:SetPoint("CENTER");
	_G.InterfaceOptions_AddCategory(frame);
end

function addon:InitQueueData()
	for name, queueCategory in pairs(self.QueueCategories) do
		if queueCategory.OnLoad then
			queueCategory:OnLoad();
		end
		self.BaseLFGCategory.OnLoad(queueCategory);
	end
end

addon.BaseLFGCategory = {
	OnLoad = function(self)
		-- Setup the internal data structures before mucking with the metatable
		self.queueData = {
			Data = {} };
		self.parent = addon.BaseLFGCategory;
		local mt = {
			__index = addon.BaseLFGCategory
		};
		setmetatable(self, mt);
	end,
	UpdateData = function(self)
		local hasData,  leaderNeeds, tankNeeds, healerNeeds, dpsNeeds, totalTanks, totalHealers, totalDPS, instanceType, instanceSubType, instanceName, averageWait, tankWait, healerWait, damageWait, myWait, queuedTime
		local mode, subMode;
		local lfgCategory = self:GetLFGCategory();
		local queueData = self:GetQueueData();
		
		wipe(queueData.Data);
		queueData.Mode = nil;
		queueData.SubMode = nil;
		
		hasData,  leaderNeeds, tankNeeds, healerNeeds, dpsNeeds, totalTanks, totalHealers, totalDPS, instanceType, instanceSubType, instanceName, averageWait, tankWait, healerWait, damageWait, myWait, queuedTime = GetLFGQueueStats(lfgCategory);
		
		queueData.Data.HaveData = hasData;
		queueData.Data.NeedLeaders = leaderNeeds or 0;
		queueData.Data.NeedTanks = tankNeeds or 0;
		queueData.Data.NeedHealers = healerNeeds or 0;
		queueData.Data.NeedDPS = dpsNeeds or 0;
		queueData.Data.TotalTanks = totalTanks or 1; -- Failsafe counts for missing dungeon info
		queueData.Data.TotalHealers = totalHealers or 1;
		queueData.Data.TotalDPS = totalDPS or 3;
		queueData.Data.QueueType = instanceType;
		queueData.Data.QueueSubType = instanceSubType;
		queueData.Data.QueueName = instanceName;
		queueData.Data.AverageWait = averageWait or -1;
		queueData.Data.TankWait = tankWait or -1;
		queueData.Data.HealerWait = healerWait or -1;
		queueData.Data.DPSWait = damageWait or -1;
		queueData.Data.MyWait = myWait or -1;
		queueData.Data.QueueTime = queuedTime;
		
		mode, subMode = GetLFGMode(lfgCategory);
		
		queueData.Mode = mode;
		queueData.SubMode = subMode;
	end,
	GetLFGCategory = function(self)
		return self.lfgCategory;
	end,
	GetQueueData = function(self)
		return self.queueData;
	end,
	GetConfigData = function(self)
		return db.Frame.Queue.Data[self.configKey];
	end,
	GetConfigKey = function(self)
		return self.configKey;
	end,
	GetStateDispatcher = function(self)
		return self.stateDispatcher;
	end,
	GetStateKey = function(self)
		local queueData = self:GetQueueData();
		local mode = queueData.Mode;
		local queueStats = queueData.Data;
		if (not mode or mode == "abandonedInDungeon") and not queueStats.HaveData then -- not queued
			return "notQueued";
		elseif mode == "proposal" then -- queue popped, waiting to join
			return "proposal";
		elseif mode == "lfgparty" then -- in an instance
			return "inParty";
		elseif (mode == "queued" or mode == "listed" or mode == "rolecheck") then
			if not queueStats.HaveData then
				return "queuedNoData";
			else
				return "queued";
			end
		else
			return "unknown";
		end
	end,
	ShouldShowFrameData = function(self)
		self:UpdateData();
		local configKey = self:GetConfigKey();
		-- Check trivial first
		-- Check if we are showing data for this queue type
		if not db.Frame.Queue.Types[configKey] then
			return false;
		end
		-- Check based on the queue state
		local stateKey = self:GetStateKey();
		-- GetStateKey can return an array of state keys
		if type(stateKey) == "table" then
			for _, key in pairs(stateKey) do
				-- At least one queue sub-type has data to show
				if db.Frame.Queue.States[key] then
					return true;
				end
			end
			return false;
		else
			if not db.Frame.Queue.States[stateKey] then
				return false;
			else
				return true;
			end
		end
	end,
	UpdateFrameData = function(self, frame)
		if self:ShouldShowFrameData() then
			local stateKey = self:GetStateKey();
			local stateDispatch = self:GetStateDispatcher();
			if stateDispatch then
				if type(stateKey) == "table" then
					for index, stateKey in pairs(stateKey) do
						if stateDispatch[stateKey] and self:ShouldShowForQueue(index) then
							stateDispatch[stateKey](self, frame, index);
						end
					end
				else
					if stateDispatch[stateKey] then
						stateDispatch[stateKey](self, frame);
					end
				end
			end
		end
	end,
}

addon.QueueCategories = {
	LookingForDungeon = {
		lfgCategory = LE_LFG_CATEGORY_LFD,
		configKey = "LookingForDungeon",
		stateDispatcher = {
			proposal = function(queueCategory, frame)
				local queueData = queueCategory:GetQueueData();
				frame:AddHeaderLine(L["Found a group"]);
				if queueData.SubMode == "accepted" then
					frame:AddHeaderLine(L["Waiting on group members"]);
				end
			end,
			inParty = function(queueCategory, frame)
				frame:AddHeaderLine(L["You are in an LFG party"]);
			end,
			queuedNoData = function(queueCategory, frame)
				local queueData = queueCategory:GetQueueData();
				frame:AddHeaderLine(L["Acquiring LFG queue data..."]);
				if queueData.Data.QueueTime then
					frame:AddHeaderLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - queueData.Data.QueueTime)));
				end
			end,
			queued = function(queueCategory, frame)
				local queueData = queueCategory:GetQueueData();
				local queueStats = queueData.Data;
				local configData = queueCategory:GetConfigData();
				if queueStats.QueueName then
					frame:AddHeaderLine((L["Dungeon type"] .. ": %s"):format(queueStats.QueueName));
				end
				if configData.QueueTimes.Tank and queueStats.TotalTanks > 0 then
					frame:AddLine((L["Tank queue times"] .. ": %s"):format(queueStats.TankWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.TankWait)));
				end
				if configData.QueueTimes.Healer and queueStats.TotalHealers > 0 then
					frame:AddLine((L["Healer queue times"] .. ": %s"):format(queueStats.HealerWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.HealerWait)));
				end
				if configData.QueueTimes.DPS then
					frame:AddLine((L["DPS queue times"] .. ": %s"):format(queueStats.DPSWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.DPSWait)));
				end
				if configData.QueueTimes.Self then
					frame:AddLine((L["Your queue time"] .. ": %s"):format(queueStats.MyWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.MyWait)));
				end
				if queueStats.QueueTime then
					frame:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - queueStats.QueueTime)));
				end
				local haveTanks = queueStats.TotalTanks - queueStats.NeedTanks;
				local haveHealers = queueStats.TotalHealers - queueStats.NeedHealers;
				local haveDPS = queueStats.TotalDPS - queueStats.NeedDPS;
				local currentPartySize = haveTanks + haveHealers + haveDPS;
				local partySize = queueStats.TotalTanks + queueStats.TotalHealers + queueStats.TotalDPS;
				if configData.Party.Tank and queueStats.TotalTanks > 0 then
					frame:AddLine((L["Tanks found"] .. ": %d/%d"):format(haveTanks, queueStats.TotalTanks));
				end
				if configData.Party.Healer and queueStats.TotalHealers > 0 then
					frame:AddLine((L["Healers found"] .. ": %d/%d"):format(haveHealers, queueStats.TotalHealers));
				end
				if configData.Party.DPS and queueStats.TotalDPS > 0 then
					frame:AddLine((L["DPS found"] .. ": %d/%d"):format(haveDPS, queueStats.TotalDPS));
				end
				if configData.Party.Total and partySize > 0 then
					frame:AddLine((L["Total found"] .. ": %d/%d"):format(currentPartySize, partySize));
				end
			end,
		},
	},
	RaidFinder = {
		lfgCategory = LE_LFG_CATEGORY_RF,
		configKey = "RaidFinder",
		OnLoad = function(self)
			self.stateDispatcher = addon.QueueCategories.LookingForDungeon.stateDispatcher;
		end
	},
	Scenarios = {
		lfgCategory = LE_LFG_CATEGORY_SCENARIO,
		configKey = "Scenarios",
		OnLoad = function(self)
			self.stateDispatcher = addon.QueueCategories.LookingForDungeon.stateDispatcher;
		end
	},
	PvP = {
		configKey = "PvP",
		stateKeys = {},
		queueData = {},
		GetQueueData = function(self, index)
			if not self.queueData[index] then
				self.queueData[index] = {};
			end
			return self.queueData[index];
		end,
		UpdateData = function(self)
			for k, v in pairs(self.queueData) do
				wipe(v);
			end
			
			local status, mapName, teamSize, registeredMatch, timeInQueue, waitTime, queueData;
			for index = 1, GetMaxBattlefieldID() do
				status, mapName, teamSize, registeredMatch = GetBattlefieldStatus(index);
				timeInQueue = GetBattlefieldTimeWaited(index) / 1000;
				waitTime = GetBattlefieldEstimatedWaitTime(index) / 1000;
				queueData = self:GetQueueData(index);
				queueData.Status = status;
				queueData.Name = mapName;
				queueData.TeamSize = teamSize;
				queueData.TimeInQueue = timeInQueue;
				queueData.WaitTime = waitTime;
			end
		end,
		ShouldShowForQueue = function(self, index)
			local stateKey = self:GetStateKeyForQueue(index);
			if not db.Frame.Queue.States[stateKey] then
				return false;
			else
				return true;
			end
		end,
		GetStateKeyForQueue = function(self, index)
			local queueData = self:GetQueueData(index);
			local status = queueData.Status;
			if status == "queued" then
				return "queued";
			elseif status == "confirm" then
				return "proposal";
			elseif status == "active" then
				return "inParty";
			else
				return "notQueued";
			end
		end,
		GetStateKey = function(self)
			local stateKey;
			wipe(self.stateKeys);
			
			for index = 1, GetMaxBattlefieldID() do
				stateKey = self:GetStateKeyForQueue(index);
				self.stateKeys[index] = stateKey;
			end
			
			return self.stateKeys;
		end,
		stateDispatcher = {
			proposal = function(queueCategory, frame, index)
				frame:AddHeaderLine("PvP queue is ready");
			end,
			inParty = function(queueCategory, frame, index)
				frame:AddHeaderLine("In a PvP area");
			end,
			queued = function(queueCategory, frame, index)
				local queueData = queueCategory:GetQueueData(index);
				local configData = queueCategory:GetConfigData();
				frame:AddHeaderLine((L["PvP queue"] .. ": %s"):format(queueData.Name));
				if configData.TeamSize and queueData.TeamSize ~= nil and queueData.TeamSize > 0 then
					frame:AddLine((L["Team size"] .. ": %d"):format(queueData.TeamSize));
				end
				if configData.WaitTime then
					frame:AddLine((L["Wait time"] .. ": %s"):format(queueData.WaitTime == 0 and L["Instant or Unknown"] or SecondsToTime(queueData.WaitTime)));
				end
				frame:AddLine((L["Time in queue"] .. ": %s"):format(SecondsToTime(queueData.TimeInQueue)));
			end,
		},
	},
	WorldPvP = {
		configKey = "WorldPvP",
		stateKeys = {},
		queueData = {},
		GetQueueData = function(self, index)
			if not self.queueData[index] then
				self.queueData[index] = {};
			end
			return self.queueData[index];
		end,
		UpdateData = function(self)
			for k, v in pairs(self.queueData) do
				wipe(v);
			end
			
			local status, mapName, queueID, timeLeft, queueData;
			for index = 1, MAX_WORLD_PVP_QUEUES do
				status, mapName, queueID, timeLeft = GetWorldPVPQueueStatus(index);
				queueData = self:GetQueueData(index);
				queueData.Status = status;
				queueData.Name = mapName;
				queueData.QueueID = queueID;
				queueData.TimeLeft = timeLeft;
			end
		end,
		ShouldShowForQueue = function(self, index)
			local stateKey = self:GetStateKeyForQueue(index);
			if not db.Frame.Queue.States[stateKey] then
				return false;
			else
				return true;
			end
		end,
		GetStateKeyForQueue = function(self, index)
			local status = self:GetQueueData(index).Status;
			if status == "queued" then
				return "queued";
			elseif status == "confirm" then
				return "proposal";
			elseif status == "active" then
				return "inParty";
			else
				return "notQueued";
			end
		end,
		GetStateKey = function(self)
			local stateKey;
			wipe(self.stateKeys);
			
			for index = 1, MAX_WORLD_PVP_QUEUES do
				stateKey = self:GetStateKeyForQueue(index);
				self.stateKeys[index] = stateKey;
			end
			
			return self.stateKeys;
		end,
		stateDispatcher = {
			proposal = function(queueCategory, frame, index)
				frame:AddHeaderLine("PvP queue is ready");
			end,
			inParty = function(queueCategory, frame, index)
				frame:AddHeaderLine("In a PvP area");
			end,
			queued = function(queueCategory, frame, index)
				local queueData = queueCategory:GetQueueData(index);
				local configData = queueCategory:GetConfigData();
				frame:AddHeaderLine((L["World PvP queue"] .. ": %s"):format(queueData.Name));
				if configData.TimeLeft then
					frame:AddLine((L["Time remaining"] .. ": %d"):format(type(queueData.TimeLeft) == "number" and SecondsToTime(queueData.TimeLeft) or L["Unknown"]));
				end
			end,
		},
	},
};

function addon:UpdateQueueData()
	for _, queueCategory in pairs(self.QueueCategories) do
		queueCategory:UpdateData();
	end
end

function addon:OpenConfig()
	if not IsAddOnLoaded("LFGTimer_Options") or not addon.Config then
		LoadAddOn("LFGTimer_Options");
	end
	
	if not addon.Config then
		print("LFGTimer: Could not open the options window. '", IsAddOnLoaded(), "'");
	else
		HideUIPanel(InterfaceOptionsFrame);
		HideUIPanel(GameMenuFrame);
		addon.Config:Show()
	end
end

function addon:UpdateFrameShown()
	if self.CloneFrame then
		self.CloneFrame:Hide();
	end
	if self.CustomFrame then
		self.CustomFrame:Hide();
	end
	
	if self:ShouldShowFrame() then
		local frame = self:GetFrame();
		frame:Show();
		--self:GetFrame():Show();
	end
end

function addon:GetStateKey(lfgData)
	if not lfgData or not lfgData.Data then
		return "notQueued";
	end
	local mode = lfgData.Mode;
	local queueStats = lfgData.Data;
	if (not mode or mode == "abandonedInDungeon") and not queueStats.HaveData then -- not queued
		return "notQueued";
	elseif mode == "proposal" then -- queue popped, waiting to join
		return "proposal";
	elseif mode == "lfgparty" then -- in an instance
		return "inParty";
	elseif (mode == "queued" or mode == "listed" or mode == "rolecheck") then
		if not queueStats.HaveData then
			return "queuedNoData";
		else
			return "queued";
		end
	else
		return "unknown";
	end
end

addon.LFGStates = {
	notQueued = {
		ShouldShow = function(lfgData)
			return db.Frame.ShowWhen.Not;
		end,
		UpdateFrameData = function(frame, lfgData)
			--frame:AddLine(L["You are not in an LFG queue"]);
		end
	},
	proposal = {
		ShouldShow = function(lfgData)
			return db.Frame.ShowWhen.Found;
		end,
		UpdateFrameData = function(frame, lfgData)
			frame:AddLine(L["Found a group"]);
			if lfgData.SubMode == "accepted" then
				frame:AddLine(L["Waiting on group members"]);
			end
		end
	},
	inParty = {
		ShouldShow = function(lfgData)
			return db.Frame.ShowWhen.Party;
		end,
		UpdateFrameData = function(frame, lfgData)
			frame:AddLine(L["You are in an LFG party"]);
		end
	},
	queuedNoData = {
		ShouldShow = function(lfgData)
			return db.Frame.ShowWhen.Acquire;
		end,
		UpdateFrameData = function(frame, lfgData)
			frame:AddLine(L["Acquiring LFG queue data..."]);
			if lfgData.Data.QueueTime then
				frame:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - lfgData.Data.QueueTime)));
			end
		end
	},
	queued = {
		ShouldShow = function(lfgData)
			return db.Frame.ShowWhen.Wait;
		end,
		UpdateFrameData = function(frame, lfgData)
			local queueStats = lfgData.Data;
			if db.Frame.Data.Type and queueStats.QueueName then
				frame:AddHeaderLine((L["Dungeon type"] .. ": %s"):format(queueStats.QueueName));
			end
			if db.Frame.Data.Tank and queueStats.TotalTanks > 0 then
				frame:AddLine((L["Tank queue times"] .. ": %s"):format(queueStats.TankWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.TankWait)));
			end
			if db.Frame.Data.Healer and queueStats.TotalHealers > 0 then
				frame:AddLine((L["Healer queue times"] .. ": %s"):format(queueStats.HealerWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.HealerWait)));
			end
			if db.Frame.Data.DPS then
				frame:AddLine((L["DPS queue times"] .. ": %s"):format(queueStats.DPSWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.DPSWait)));
			end
			if db.Frame.Data.Self then
				frame:AddLine((L["Your queue time"] .. ": %s"):format(queueStats.MyWait == -1 and L["Instant or Unknown"] or SecondsToTime(queueStats.MyWait)));
			end
			if queueStats.QueueTime then
				frame:AddLine((L["Time in queue: %s"]):format(SecondsToTime(GetTime() - queueStats.QueueTime)));
			end
			if db.Frame.Data.Party then
				local haveTanks, haveHealers, haveDPS = queueStats.TotalTanks - queueStats.NeedTanks, queueStats.TotalHealers - queueStats.NeedHealers, queueStats.TotalDPS - queueStats.NeedDPS;
				local totalHave, totalNeed = haveTanks + haveHealers + haveDPS, queueStats.TotalTanks + queueStats.TotalHealers + queueStats.TotalDPS;
				if db.Frame.Data.Found.Tank and queueStats.TotalTanks > 0 then
					frame:AddLine((L["Tanks found"] .. ": %d/%d"):format(haveTanks, queueStats.TotalTanks));
				end
				if db.Frame.Data.Found.Healer and queueStats.TotalHealers > 0 then
					frame:AddLine((L["Healers found"] .. ": %d/%d"):format(haveHealers, queueStats.TotalHealers));
				end
				if db.Frame.Data.Found.DPS and queueStats.TotalDPS > 0 then
					frame:AddLine((L["DPS found"] .. ": %d/%d"):format(haveDPS, queueStats.TotalDPS));
				end
				if db.Frame.Data.Found.Total and totalNeed > 0 then
					frame:AddLine((L["Party members found"] .. ": %d/%d"):format(totalHave, totalNeed));
				end
			end
		end
	},
	unknown = {
		ShouldShow = function(lfgData)
			return false;
		end,
		UpdateFrameData = function(frame, lfgData)
			frame:AddLine(L["Found an odd state."]);
			frame:AddLine(L["Waiting on updated data from the server."]);
		end
	},
};

function addon:ShouldShowFrame()
	-- Trivial case
	if not db.Frame.Show then
		return false;
	end
	
	local stateKey;
	
	--self:UpdateQueueData();
	
	for k, queueCategory in pairs(self.QueueCategories) do
		if queueCategory:ShouldShowFrameData() then
			return true;
		end
	end
	
	return false;
end

function addon:GetFrame()
	if db.Frame.Type == "clone" then
		return self:SetupCloneFrame();
	elseif db.Frame.Type == "custom" then
		return self:SetupCustomFrame();
	end
end

function addon:IsFrameLocked()
	return db.FrameLocked;
end

function addon:SetAnchor(frame)
	local anchor = db.Frame.Anchor;
	frame:ClearAllPoints();
	frame:SetPoint(anchor.Point, anchor.To, anchor.At, anchor.X, anchor.Y);
end

function addon:UpdateFramePosition(frame)
	local anchor = db.Frame.Anchor;
	local point, to, at, x, y = frame:GetPoint(1);
	anchor.Point = point;
	anchor.At = at;
	anchor.X = x;
	anchor.Y = y;
end

function addon:SetupCloneFrame()
	-- Create clone frame
	local frame, f; -- Declare variables
	
	frame = CreateFrame("Frame", "LFGTimerQueueStatusFrame", UIParent, "TooltipBorderedFrameTemplate");
	frame:Hide();
	frame:SetMovable(true);
	frame:SetFrameStrata("TOOLTIP");
	frame:SetClampedToScreen(true);
	frame:SetHeight(150);
	frame:SetWidth(275);
	frame:EnableMouse(true);
	frame:SetUserPlaced(true);
	self:SetAnchor(frame);
	
	QueueStatusFrame_OnLoad(frame);
	frame:SetScript("OnEvent", QueueStatusFrame_OnEvent);
	
	-- Setup movement
	frame:SetScript("OnMouseDown", function(frame, button)
		if (not addon:IsFrameLocked()) then
			frame:StartMoving();
		end
	end);
	frame:SetScript("OnMouseUp", function(frame, button)
		frame:StopMovingOrSizing();
		addon:UpdateFramePosition(frame);
	end);
	
	self.SetupCloneFrame = function(self)
		return self.CloneFrame;
	end -- Overwrite the setup function to just return the frame
	
	self.CloneFrame = frame;
	
	return frame;
end

function addon:SetupCustomFrame()
	local frame = CreateFrame("Frame", "LFGTimerCustomTooltipFrame", UIParent, "LFGTimerCustomTooltipFrameTemplate");
	
	self:SetAnchor(frame);

	self.CustomFrame = frame;
	
	self.SetupCustomFrame = function(self)
		return self.CustomFrame;
	end -- Overwrite the setup function to just return the frame
	
	return frame;
end

function addon:UpdateFrameArt()
	if self.CustomFrame then
		self.CustomFrame:UpdateFrameArt();
	end
end

addon.CustomFrameFuncs = {
	OnLoad = function(frame)
		for k, v in pairs(addon.CustomFrameFuncs) do
			frame[k] = v;
		end
		
		frame.Lines = {};
		frame.TextFrames = {};
		frame.NumTextFrames = 0;
		frame.Backdrop = {
			insets = {},
		};

		frame:SetScript("OnShow", frame.OnShow);
		frame:SetScript("OnHide", frame.OnHide);
		frame:SetScript("OnMouseDown", frame.OnMouseDown);
		frame:SetScript("OnMouseUp", frame.OnMouseUp);
		frame:UpdateFrameArt();
	end,
	OnShow = function(frame)
		addon:SetAnchor(frame)
		frame:SetScript("OnUpdate", frame.UpdateData);
	end,
	OnHide = function(frame)
		frame:SetScript("OnUpdate", nil);
	end,
	OnMouseDown = function(frame, button)
		if not addon:IsFrameLocked() then
			frame:StartMoving();
		end
	end,
	OnMouseUp = function(frame)
		frame:StopMovingOrSizing();
		addon:UpdateFramePosition(frame);
	end,
	UpdateFrameArt = function(frame)
		local inset = db.Frame.Backdrop.Inset;
		local backdrop = frame.Backdrop;
		
		backdrop.bgFile = LSM:Fetch("background", db.Frame.Backdrop.Background);
		backdrop.edgeFile = LSM:Fetch("border", db.Frame.Backdrop.Border);
		backdrop.tile = db.Frame.Backdrop.Tile;
		backdrop.tileSize = db.Frame.Backdrop.TileSize;
		backdrop.edgeSize = db.Frame.Backdrop.EdgeSize;
		
		backdrop.insets.left = inset;
		backdrop.insets.right = inset;
		backdrop.insets.top = inset;
		backdrop.insets.bottom = inset;
		
		local bgColor, borderColor = db.Frame.Backdrop.BackgroundColor, db.Frame.Backdrop.BorderColor;
		
		frame:SetBackdrop(backdrop);
		frame:SetBackdropColor(bgColor.r, bgColor.g, bgColor.b, bgColor.a);
		frame:SetBackdropBorderColor(borderColor.r, borderColor.g, borderColor.b, borderColor.a);
		
		frame.Container:ClearAllPoints();
		frame.Container:SetPoint("TOP", 0, -inset);
		frame.Container:SetPoint("BOTTOM", 0, inset);
		frame.Container:SetPoint("LEFT", inset, 0);
		frame.Container:SetPoint("RIGHT", -inset, 0);
		
		frame:UpdateSize();
	end,
	ClearLines = function(frame)
		for k, v in pairs(frame.Lines) do
			frame.Lines[k] = nil;
			v:Hide();
			v.Text:SetText("Empty");
			frame.TextFrames[v] = true;
		end
	end,
	GetTextFrame = function(frame)
		local f = next(frame.TextFrames);
		if not f then
			frame.NumTextFrames = frame.NumTextFrames + 1;
			f = CreateFrame("Frame", "LFGTimerTextFrame" .. frame.NumTextFrames, frame.Container, "LFGTimerTextFrameTemplate");
			for k, v in pairs(addon.LineFrameFuncs) do
				f[k] = v;
			end
		else
			frame.TextFrames[f] = nil;
		end
		
		local font, size, flags, color = (LSM:Fetch("font", db.Frame.Text.Font)), db.Frame.Text.Size, db.Frame.Text.Outline, db.Frame.Text.Color;
		
		f.Text:SetText("Initialized");
		f.Text:Show();
		f.Text:SetFont(font, size, flags);
		f.Text:SetTextColor(color.r, color.g, color.b, color.a);
		return f;
	end,
	AddLine = function(frame, text)
		local f = frame:GetTextFrame();
		f.Text:SetText(text);
		f:SetIsHeader(false);
		table.insert(frame.Lines, f);
	end,
	AddHeaderLine = function(frame, text)
		local f = frame:GetTextFrame();
		f.Text:SetText(text);
		f:SetIsHeader(true);
		table.insert(frame.Lines, f);
	end,
	UpdateSize = function(frame)
		local inset = db.Frame.Backdrop.Inset;
		local last = nil;
		local maxWidth, height = 0, (2 * inset) + 10;
		local headerLineWidthAdjust = 0;
		
		for i, v in ipairs(frame.Lines) do
			headerLineWidthAdjust = 0;
			v:Show();
			v.Text:Show();
			v:ClearAllPoints();
			if not last then
				v:SetPoint("TOP", frame.Container, "TOP", 0, -inset);
			else
				v:SetPoint("TOP", last, "BOTTOM", 0, 0);
			end
			if not v:GetIsHeader() then
				headerLineWidthAdjust = 5;
			end
			v:SetPoint("LEFT", frame.Container, "LEFT", inset + headerLineWidthAdjust, 0);
			v:SetWidth(v.Text:GetStringWidth() + (2 * inset) + headerLineWidthAdjust);
			v:SetHeight(v.Text:GetStringHeight());
			height = height + v:GetHeight();
			
			if v:GetWidth() > maxWidth then
				maxWidth = v:GetWidth();
			end
			last = v;
		end
		
		frame:SetWidth(maxWidth + (2 * inset));
		frame:SetHeight(height);
	end,
	UpdateData = function(frame)
		frame:ClearLines();
		
		for _, queueCategory in pairs(addon.QueueCategories) do
			if queueCategory:ShouldShowFrameData() then
				queueCategory:UpdateFrameData(frame);
			end
		end
		
		if #frame.Lines == 0 then
			frame:AddHeaderLine(L["You are not in a queue"]);
		end
		
		frame:UpdateSize();
	end,
};

addon.LineFrameFuncs = {
	SetFormattedText = function(frame, format, ...)
		local text = addon:Format(format, ...);
		if not text then
			text = "";
		end
		frame.Text:SetText(text);
	end,
	SetText = function(frame, text)
		if not text then
			text = "";
		end
		frame.Text:SetText(text);
	end,
	SetIsHeader = function(frame, flag)
		frame.IsHeader = flag;
	end,
	GetIsHeader = function(frame)
		return frame.IsHeader;
	end
};
